
#### HAND-WRITTEN PARSER

def parse(text) :
    """driver for parser
    
       pre:  text  is a string, representing a proposition-logic expression
       post: answer is a nested-list representation of a parse tree for  text
       returns: ans
    """
    global wordlist
    wordlist = scan(text)
    getNextword()   # initialize parser
    answer = parseEXPR()
    return answer


##### SCANNING : converting a string into a list of words ################

EOF = "!"      # a word that marks the end of the input words
nextword = ""  # holds the next unread input word
wordlist = []  # holds remaining input words
# global invariant:  nextword + wordlist == all remaining unread input words


def scan(text) :
    """scan splits apart the symbols/words in  text  into a list of words.

       pre:  text is a string holding a proposition
       post: answer is a wordlist of the words and symbols within  text
             (spaces and newlines are removed)
       returns: answer
    """
    OPERATORS = ("(", ")", "^", "v", "~")  # must be one-symbol operators
    SPACES = (" ", "\n")
    SEPARATORS = OPERATORS + SPACES
 
    next = ""  # the next symbol/word to be added to the answer list
    answer = []

    for letter in text:
        # invariant:  answer + next + letter  equals all the words/symbols
        #     read so far from  text  with  SPACES  removed

        # see if  next  is complete and should be appended to answer:
        if letter in SEPARATORS  and  next != "" :
            answer.append(next)
            next = ""

        if letter in OPERATORS :
            answer.append(letter)
        elif letter in SPACES :
            pass  # discard space
        else :    # build a word or numeral
            next = next + letter

    if next != "" :   
        answer.append(next)

    return answer


def getNextword() :
    """moves the front word in  wordlist  to  nextword.
       If wordlist is empty, sets  nextword = EOF
    """
    global nextword, wordlist
    if wordlist == [] :
        nextword = EOF
    else :
        nextword = wordlist[0]
        wordlist = wordlist[1:]

        

####  PARSING :  translating a list of words (strings) into a tree   ######

def parseEXPR() :
    """builds an EXPR operator tree from the words in  nextword + wordlist,
          where  EXPR ::=  VAR  |  ( EXPR OP EXPR )  | ~ EXPR
                 OP is  "^"  or  "v"
                 VAR is an alphabetic letter

       pre: global var  nextword  holds the next unread word
       post:  ans  is an operator tree holding the expression that
              was read starting from  nextword.
       returns: ans
    """
    def convert(operator) :
        """helper function that converts operator symbol to a token"""
        if operator == "^" : return "and"
        elif operator == "v": return "or"
        else: return "ERROR"
            
    if  nextword.isalpha() :   # a VAR
        ans = nextword        
        getNextword()
    elif nextword == "(" :     # ( EXPR op EXPR ) ?
        getNextword()
        tree1 = parseEXPR()
        op = nextword
        if op == "^"  or  op == "v" :
            getNextword()
            tree2 = parseEXPR()
            if nextword == ")" :
                ans = [convert(op), tree1, tree2]
                getNextword()
            else :
                error("missing )")
        else :
            error("missing operator")
    elif nextword == "~" :
        getNextword()
        tree = parseEXPR()
        ans = ["not", tree]
    else :
        print nextword, wordlist
        error("illegal symbol to start an expression")

    return ans


###### TOSTRING (FORMATTING) FUNCTION:

def toString(tree):
    """reformats  tree  as a linear string

       pre: tree is an optree
       post: ans is a string that holds the words within tree
       returns: ans
    """
    
    def unconvert(op) :
        if op == "or" : return "v"
        elif op == "and" : return "^"
        else: return "ERROR"

    if isinstance(tree, str) :  # a VAR ?
        ans = tree
    else :  # list of form,  [op, ...]  where  op in ("~","^","v")
        if tree[0] == "not" :
            ans = "~" + toString(tree[1])
        elif tree[0] in ("or", "and") :
            ans = "(" + toString(tree[1]) + " " + unconvert(tree[0]) + " " + toString(tree[2]) + ")"
        else :
            error("bad optree structure")
            ans = "?"
    return ans


def error(message) :
    print message
