
# PLY template of parser for baby logic language

from ply import *
import lexPLY

tokens = lexPLY.tokens

# GRAMMAR RULES and their output formats:
# A PLY-coded grammar must begin with a rule for a 'Program':

#   P ::=  E1 OP E2  |  E     (no requirement for parens around entire expression)
def p_Program1(p):
    '''Program : Expression Op Expression '''
    p[0] = [p[2], p[1], p[3]]

def p_Program2(p):
    '''Program : Expression''' 
    p[0] = p[1]
    
#    E ::=  ID  |  ( E1 OP E2 )  |  ~ E  |  True  |  False
#    ETREE ::=  ID |  [BOPP, ETREE, ETREE] |  ["not", ETREE]  |  "True"  |  "False"
def p_Expression1(e):
    '''Expression : ID'''
    e[0] = e[1] 

def p_Expression2(e):
    '''Expression : '(' Expression Op Expression ')' '''
    e[0] = [e[3], e[2], e[4]]

def p_Expression3(e):
    '''Expression : '~' Expression '''
    e[0] = ["not", e[2]]
    
def p_Expression4(e):
    '''Expression : True '''
    e[0] = "True"

def p_Expression5(e):
    '''Expression : False '''
    e[0] = "False"

#    OP ::=  ^  |  v
#    BOP  ::=  "and"  |  "or"
def p_Op1(o):
    '''Op : '^' '''
    o[0] = "and"

def p_Op2(o):
    '''Op : 'v' '''
    o[0] = "or"


#### Error handler : prints location of error and quits parser
def p_error(p):
    print "SYNTAX ERROR at LexToken(grammar_phrase, input_word, line_number, char_number):"
    print (15 * " "), p
    print "PARSER QUITS."
    raise Exception

#### Driver:
bparser = yacc.yacc()

def parse(data,debug=0):
    bparser.error = 0
    p = bparser.parse(data,debug=debug)
    if bparser.error: return None
    return p

