
""""Translate.py   translates a proposition written with AND,OR,NOT so that
    all occurrences of AND are replaced by OR,NOT and then simplifies.

    Input: a parse tree (nested-list representation) whose  structure is defined:
    
            ETREE ::=  [BOP, ETREE, ETREE ]  | ["not", ETREE]  
                     |  "True" | "False"  | ID
            BOP ::=  "and"  |  "or"

    Output: a tree, in the above format, that contains no  "and"  operations and
            whose (truth-table) meaning is identical to the meaning
            of the Input.

    Algorithm:
     0. Receive the input tree.
      
     1. Replace each occurrence of  ["and", t1, t2]
        by  ["not", ["or", ["not", t1], ["not", t2]]]

     2. while it is still possible to do so:
           (i) Replace double negations,  ["not", ["not", t]],  by just  t.
     
          (ii) Replace occurrences of ["or", "True", t]  by  "True"
                                  ["or", t, "True"]  by  "True"
                                  ["or", "False", t]  by  t
                                  ["or", t, "False"]  by  t
                                
         (iii) Replace occurrences of ["not", "True"]  by  "False"
                                      ["not", "False"]  by  "True"
"""

###### TOSTRING (PRETTY-PRINT-FORMATTING) FUNCTION:

def toString(tree):
    """reformats  tree  as a linear string

       pre: tree is an ETREE
       post: ans is the string representation of  tree
       returns: ans
    """
    
    def unconvert(op) :    # helper function
        if op == "or" : return "v"
        elif op == "and" : return "^"
        else: return "ERROR"

    if isinstance(tree, str) :  # ID/True/False ?
        ans = tree
    else :  # [op, ...]  where  op in ("not", "and", "or")
        if tree[0] == "not" :
            ans = "~" + toString(tree[1])
        elif tree[0] in ("or", "and") :
            ans = "(" + toString(tree[1]) + " " + unconvert(tree[0]) + " " + toString(tree[2]) + ")"
        else :
            error("bad tree structure")
            ans = "ERROR"
    return ans



###### TRANSLATION AND SIMPLIFICAIONT FUNCTIONS:  ########

def removeAnd(tree) :
    # DOCUMENT ME AND WRITE ME
    return tree


def removeDoubleNeg(tree):
    # DOCUMENT ME AND WRITE ME
    return tree


def optimizeOr(tree):
    """simplifies one or more of these combinations of "or" embedded in a parse tree:
            ["or", "True", t]   ==>  "True"
            ["or", t, "True"]   ==>  "True"
            ["or", "False", t]  ==>   t
            ["or", t, "False"]  ==>   t        
       pre:  tree  is a parse tree
       post: ans  is  tree  simplified as stated above
       returns: ans
    """  
    if isinstance(tree, str) :   # an ID/True/False
        ans = tree
    else :   #  tree == [op, ...]
        op = tree[0]
        if op == "not" :
            # no work to do at the top of the expression, but maybe
            # there is something to do within the subexpression:
            ans = ["not", optimizeOr(tree[1])]
        elif op == "or" :
            # see if there is a top-level optimization to do:
            if (tree[1] == "True") or (tree[2] == "True") :
                ans = "True"
            elif (tree[1] == "False") or (tree[2] == "False") :
                ans = tree # REPLACE THIS LINE WITH SOMETHING BETTER
            else :
                # maybe there is something to do in a subexpression:
                ans = tree # REPLACE THIS LINE WITH SOMETHING BETTER
        else :  # some other logical operator we don't know???
            ans = tree
    return ans
    
    
def optimizeNeg(tree):
    # DOCUMENT ME AND WRITE ME:
    return tree


#### TABLE of optimizations:  #########
"""  optimization : { LABEL : TREEFUNCTION },
        where  LABEL = string
               TREEFUNCTION = ETREE -> ETREE
"""
optimization = { "double negation removed:" : removeDoubleNeg,
                 "or removed:" : optimizeOr,
                 "not removed:" : optimizeNeg
                } 
 

#### DRIVER starts here: ########################

def translate(tree) :
    """translates and optimizes  tree
        pre:  tree  is a parse tree (ETREE) that matches the grammar at the top of this file
        post: newTree  is an ETREE that has the same logical meaning as tree, 
              but optimized to eliminate "and" and to remove extra "not"s, "or"s, "True"s, and "False"s
        returns:  newTree
    """
    tree1 = removeAnd(tree)
    print "ands removed:", tree1
    print toString(tree1)

    # Optimize the translated expression:
    improving = True  # holds True when more optimizations are possible
    while improving :
        # invariant: tree1  has the same logical meaning of  tree, but it is optimized
        improving = False  # assume no more optimizations and see what happens:
        for opt in optimization.keys() :
            # applying each of the optimizations in the table, in turn:
            tree2 = optimization[opt](tree1)
            if tree1 != tree2 :
                improving = True
                print opt, tree2
                print toString(tree2)
                tree1 = tree2
    
    # all meaning-preserving optimizations are completed
    return tree1
    

    
